unit SoundPlayer;

interface

uses
  Windows, Messages, SysUtils, Classes, MMSystem;

type
  TMMErrorEvent = procedure(Sender: TObject; MMError: MMRESULT) of object;

type
  TSoundPlayer = class(TObject)
  private
    FVolume: DWORD;
    FVolumeSeted: Boolean;
    FOpened: Boolean;
    FhWaveOut: HWAVEOUT;
    FWaveRecordID: UINT;
    FWindowHandle: HWND;
    FWaveFormatEx: TWaveFormatEx;
    FBufIsFull: TNotifyEvent;
    FOnMMError: TMMErrorEvent;
    FInPause: Boolean;
    FLastError: MMRESULT;
    procedure SetOpened(const Value: Boolean);
    procedure WndProc(var Msg: TMessage);
    procedure BufIsFull(const lpWaveHdr: PWaveHdr);
    procedure SetVolume(const Value: DWORD);
  protected
    { Protected declarations }
  public
    constructor Create;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    procedure Stop;
    procedure Pause;
    procedure AddBufToPlay(lpBuf: Pointer; Count: DWORD);
    procedure SetWaveFormat(const Channels: Word; const Bits: Word; const Samples: DWORD);
    property WaveOutHandle: HWAVEOUT read FhWaveOut;
    property WaveFormatEx: TWaveFormatEx read FWaveFormatEx;
    property Opened: Boolean read FOpened write SetOpened;
    property WaveRecordID: UINT read FWaveRecordID write FWaveRecordID;
    property OnBufIsFull: TNotifyEvent read FBufIsFull write FBufIsFull;
    property OnMMError: TMMErrorEvent read FOnMMError write FOnMMError;
    property LastError: MMRESULT read FLastError;
    property Volume: DWORD read FVolume write SetVolume;
  end;

implementation

constructor TSoundPlayer.Create;
  var   WFX: TWaveFormatEx;
    hOutTmp: HWAVEOUT;
     Volume: DWORD;
begin
  inherited;
  FWindowHandle := AllocateHWnd(WndProc);
  FWaveRecordID := WAVE_MAPPER;
  SetWaveFormat(1, 8, 11025);

  FVolume := $FFFFFFFF;
  with WFX do
  begin
    wFormatTag := WAVE_FORMAT_PCM;
    nChannels := 2;
    nSamplesPerSec := 11025;
    wBitsPerSample := 16;
    nBlockAlign := nChannels * wBitsPerSample div 8;
    nAvgBytesPerSec := nBlockAlign * nSamplesPerSec;
    cbSize := 0;
  end;
  if waveOutOpen(@hOutTmp, FWaveRecordID, @WFX, 0, 0, WAVE_ALLOWSYNC or CALLBACK_NULL) = MMSYSERR_NOERROR then
  begin
    if waveOutGetVolume(hOutTmp, @Volume) = MMSYSERR_NOERROR then FVolume := Volume;
    waveOutClose(hOutTmp);
  end;
end;

destructor TSoundPlayer.Destroy;
begin
  Close;
  DeallocateHWnd(FWindowHandle);
  inherited;
end;

procedure TSoundPlayer.Open;
begin
  if not FOpened then
  begin
    FLastError := waveOutOpen(@FhWaveOut, FWaveRecordID, @FWaveFormatEx,
      FWindowHandle, 0, WAVE_ALLOWSYNC or CALLBACK_WINDOW);
    if FLastError = MMSYSERR_NOERROR then
    begin
      FOpened := True;
      if FVolumeSeted then waveOutSetVolume(FhWaveOut, FVolume);
    end else
      if Assigned(FOnMMError) then FOnMMError(Self, FLastError);
  end;
end;

procedure TSoundPlayer.Close;
begin
  if FOpened then
  begin
    FOpened := False;
    Stop;
    waveOutClose(FhWaveOut);
  end;
end;

procedure TSoundPlayer.Stop;
  var Msg: TMsg;
begin
  FInPause := False;
  FLastError := waveOutReset(FhWaveOut);
  while PeekMessage(Msg, FWindowHandle, WOM_DONE, WOM_DONE, PM_REMOVE) do
  begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
  if (FLastError <> MMSYSERR_NOERROR) and Assigned(FOnMMError) then
    FOnMMError(Self, FLastError);
end;

procedure TSoundPlayer.Pause;
begin
  if FOpened then
  begin
    if FInPause then
    begin
      FLastError := waveOutRestart(FhWaveOut);
      FInPause := FLastError <> MMSYSERR_NOERROR;
    end else
    begin
      FLastError := waveOutPause(FhWaveOut);
      FInPause := FLastError = MMSYSERR_NOERROR;
    end;
    if (FLastError <> MMSYSERR_NOERROR) and Assigned(FOnMMError) then
      FOnMMError(Self, FLastError);
  end;
end;

procedure TSoundPlayer.AddBufToPlay(lpBuf: Pointer; Count: DWORD);
  var lpWaveHdr: PWaveHdr;
begin
  if FOpened then
  begin
    New(lpWaveHdr);
    lpWaveHdr.dwFlags := 0;
    lpWaveHdr.dwBufferLength := Count;
    GetMem(lpWaveHdr.lpData, lpWaveHdr.dwBufferLength);
    Move(lpBuf^, lpWaveHdr.lpData^, Count);
    FLastError := waveOutPrepareHeader(FhWaveOut, lpWaveHdr, SizeOf(TWaveHdr));
    if FLastError = MMSYSERR_NOERROR then
      FLastError := waveOutWrite(FhWaveOut, lpWaveHdr, SizeOf(TWaveHdr));
    if FLastError <> MMSYSERR_NOERROR then
    begin
      if lpWaveHdr^.dwFlags and WHDR_PREPARED = WHDR_PREPARED then
        waveOutUnPrepareHeader(FhWaveOut, lpWaveHdr, SizeOf(TWaveHdr));
      FreeMem(lpWaveHdr.lpData, lpWaveHdr.dwBufferLength);
      Dispose(lpWaveHdr);
      if Assigned(FOnMMError) then FOnMMError(Self, FLastError);
    end;
  end;
end;

procedure TSoundPlayer.WndProc(var Msg: TMessage);
begin
  with Msg do Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
  if Msg.Msg = WOM_DONE then BufIsFull(PWaveHdr(Msg.LParam))
end;

procedure TSoundPlayer.BufIsFull(const lpWaveHdr: PWaveHdr);
begin
  if lpWaveHdr^.dwFlags and WHDR_PREPARED = WHDR_PREPARED then
    waveOutUnPrepareHeader(FhWaveOut, lpWaveHdr, SizeOf(TWaveHdr));
  FreeMem(lpWaveHdr^.lpData, lpWaveHdr^.dwBufferLength);
  Dispose(lpWaveHdr);
  if Assigned(FBufIsFull) then FBufIsFull(Self);
end;

procedure TSoundPlayer.SetWaveFormat(const Channels, Bits: Word;
  const Samples: DWORD);
  var wfx: TWaveFormatEx;
begin
  with wfx do
  begin
    wFormatTag := WAVE_FORMAT_PCM;
    nChannels := Channels;
    nSamplesPerSec := Samples;
    wBitsPerSample := Bits;
    nBlockAlign := wBitsPerSample * nChannels div 8;
    nAvgBytesPerSec := nBlockAlign * nSamplesPerSec;
    cbSize := 0;
  end;
  if not CompareMem(@wfx, @FWaveFormatEx, SizeOf(TWaveFormatEx)) then
  begin
    Close;
    FWaveFormatEx := wfx;
  end;
end;

procedure TSoundPlayer.SetOpened(const Value: Boolean);
begin
  if FOpened <> Value then if Value then Open else Close;
end;

procedure TSoundPlayer.SetVolume(const Value: DWORD);
begin
  FVolumeSeted := True;
  if FVolume <> Value then
  begin
    FVolume := Value;
    if FOpened then waveOutSetVolume(FhWaveOut, FVolume);
  end;
end;

end.
